#include <Node.h>
#include <NodeInfo.h>
#include <File.h>
#include <TypeConstants.h>
#include <String.h>
#include <UTF8.h>
#include <stdio.h>
#include <memory.h>
#include <alloc.h>
#include <fs_attr.h>
#include "bfs.h"
//#define bool int
//#define NO_TYPEDEF_BOOL
/*extern "C" {
#include "shared/node4d.h"
#include "shared/jblist.h"
#include "memmessage.h"
long FidoToTime(uchar *date);
}*/



struct bfs_attribs bfs_attribs[]=
{
	{"FTN:Area"		,B_STRING_TYPE},
	{"FTN:Status"	,B_STRING_TYPE},
	{"FTN:From"		,B_STRING_TYPE},
	{"FTN:To"		,B_STRING_TYPE},
	{"FTN:OAddr"	,B_STRING_TYPE},
	{"FTN:DAddr"	,B_STRING_TYPE},
	{"FTN:Subject"	,B_STRING_TYPE},
	{"FTN:Written"	,B_TIME_TYPE},
	{"FTN:Received"	,B_TIME_TYPE},
	{"ATTR"			,B_UINT32_TYPE}, //??? or may be 16 ?
	{"KLUDGES"		,B_MESSAGE_TYPE},
	{"MSGID"	,B_STRING_TYPE},
	{"REPLY"	,B_STRING_TYPE},
	{"FLAGS"	,B_STRING_TYPE},
	{"PATH"		,B_STRING_TYPE},
	{"SEEN-BY"	,B_STRING_TYPE},
	{"@"		,B_STRING_TYPE},
	{"LINKS"		,B_MESSAGE_TYPE},
	{"MIDCRC"	,B_UINT32_TYPE},
	{"RplCRC"	,B_UINT32_TYPE},
	{"RplTso"	,B_UINT32_TYPE},
	{"Rpl1st"	,B_UINT32_TYPE},
	{"RplNxt"	,B_UINT32_TYPE},
	{"HighWater",B_UINT32_TYPE}
};

#define HANDLE(pMsg) (long) (pMsg)
#define UNHANDLE(handle) (BMessage *) (handle)

int _bfs_ReadAttr(const char *path, int nAttr, void *pbuf, int buflen)
{
	BNode node(path);
	
	if(node.InitCheck() != B_OK)
	{
		printf("_bfs_ReadAttr: Error node retrieving - %s\r\n",path);
		return 0;
	}
	
	if(!buflen)
	{
		attr_info ai;
		node.GetAttrInfo(bfs_attribs[nAttr].name, &ai);
		return ai.size;
	}
	
	return node.ReadAttr( bfs_attribs[nAttr].name, bfs_attribs[nAttr].type, 0, pbuf, buflen);
}

int _bfs_WriteAttr(const char *path, int nAttr, void *pdata, int datalen)
{
	BNode node(path);
	
	if(node.InitCheck() != B_OK)
	{
		printf("_bfs_WriteAttr: Error node retrieving - %s\r\n",path);
		return 0;
	}
	
	return node.WriteAttr( bfs_attribs[nAttr].name, bfs_attribs[nAttr].type, 0, pdata, datalen);
}

long _bfs_CreateKludges()
{
	return HANDLE(new BMessage());
}

int _bfs_CountKludges(long kHandle, int nKludge)
{
	BMessage *pMsg = UNHANDLE(kHandle);
	int32 count;
	type_code type;
	
	pMsg->GetInfo(bfs_attribs[nKludge].name, &type, &count);

	return count;
}

int _bfs_ReadKludge(long kHandle, int nKludge, int nIdx, const void **ppbuf, int *pbuflen)
{
	BMessage *pMsg = UNHANDLE(kHandle);
	
	if( B_OK == pMsg->FindData(bfs_attribs[nKludge].name, bfs_attribs[nKludge].type, (long) nIdx, ppbuf, (ssize_t *)pbuflen))
		return 1;
	return 0;
}
int _bfs_AddKludge(long kHandle, int nKludge, const void *pdata, int datalen)
{
	BMessage *pMsg = UNHANDLE(kHandle);
	
	if( B_OK == pMsg->AddData(bfs_attribs[nKludge].name, bfs_attribs[nKludge].type, pdata, datalen, false))
		return 1;
	return 0;
}

void _bfs_ReleaseKludges(long kHandle)
{
	BMessage *pMsg = UNHANDLE(kHandle);
	delete pMsg;
}

int _bfs_ReadAttrKludges(const char *path, int nAttr, long *pkHandle)
{
	int nBufLen = _bfs_ReadAttr(path, nAttr, 0, 0);
	
	char *buff = (char *) malloc(nBufLen);
	
	int nRet = 0;
	
	if(_bfs_ReadAttr(path, nAttr, buff, nBufLen) == nBufLen)
	{
		BMessage *pMsg = new BMessage();
		if(pMsg->Unflatten(buff) == B_OK)
		{
			*pkHandle = HANDLE(pMsg);
			nRet = 1;
		}
		else
			printf("_bfs_ReadAttrKludges: Error unflattening - %s\r\n",path);
	}
	else
		printf("_bfs_ReadAttrKludges: Error attribute reading - %s\r\n",path);
	
	free(buff);

	return nRet;
}

int _bfs_WriteAttrKludges(const char *path, int nAttr, long kHandle)
{
	BMessage *pMsg = UNHANDLE(kHandle);
	 
	int nRet =0;
	ssize_t size = pMsg->FlattenedSize();
	char *buff = (char *)malloc(size);
	
	if(pMsg->Flatten(buff, size) == B_OK)
	{
		if(_bfs_WriteAttr(path, nAttr, buff, size) == size )
			nRet = 1;
	}
	else
		printf("_bfs_WriteAttrKludges: Error message flattening - %s\r\n",path);
	
	free(buff);
	return nRet;
}

int _bfs_ConvertToUTF8(unsigned long srcEncoding, const char *src, long *srcLen,
																									char *dst, long *dstLen)
{
	return convert_to_utf8(srcEncoding,src,srcLen,dst,dstLen,0);
}

int _bfs_ConvertFromUTF8(unsigned long dstEncoding, const char *src, long *srcLen,
																									char *dst, long *dstLen)
{
	return convert_from_utf8(dstEncoding, src, srcLen, dst, dstLen, 0);
}


void _bfs_SetMIMEType(char *file, char *type)
{
	BNode node(file);
	
	if(node.InitCheck() != B_OK)
		printf("_bfs_SetMIMEType: Error node retrieving - %s\r\n",file);
	
	BNodeInfo nodeInfo(&node);
	nodeInfo.SetType(type);
}
