#include <cstdio>
#include <Errors.h> 
#include <sys/stat.h> 
#include <errno.h> 
#include <dirent.h> 
#include <fcntl.h> 
#include <unistd.h> 
#include <alloc.h> 
#include "bfs.h" 
#include "storedmsg.h" 
#include "node4d.h"

#define VERSION "0.1.0"

/* Header flags */
#define FLAG_PVT             1
#define FLAG_CRASH           2
#define FLAG_RECD            4
#define FLAG_SENT            8
#define FLAG_FILEATTACH     16
#define FLAG_INTRANSIT      32
#define FLAG_ORPHAN         64
#define FLAG_KILLSENT      128
#define FLAG_LOCAL         256
#define FLAG_HOLD          512
/*      FLAG_UNUSED       1024 */
#define FLAG_FILEREQ      2048
#define FLAG_RREQ         4096
#define FLAG_IRRR         8192
#define FLAG_AUDIT       16384
#define FLAG_UPDATEREQ   32768

struct flag
{
   char *name;
   ulong jamflagbit;
   ulong fidoflagbit;
};

struct flag flagarray[] =
{ { "PVT", MSG_PRIVATE,     FLAG_PVT         },
  { "HLD", MSG_HOLD,        FLAG_HOLD        },
  { "CRA", MSG_CRASH,       FLAG_CRASH       },
  { "K/S", MSG_KILLSENT,    FLAG_KILLSENT    },
  { "SNT", MSG_SENT,        FLAG_SENT        },
  { "RCV", MSG_READ,        FLAG_RECD        },
  { "A/S", MSG_ARCHIVESENT, 0,               },
  { "DIR", MSG_DIRECT,      0                },
  { "FIL", MSG_FILEATTACH,  FLAG_FILEATTACH  },
  { "FRQ", MSG_FILEREQUEST, FLAG_FILEREQ     },
  { "IMM", MSG_IMMEDIATE,   0                },
  { "KFS", MSG_KILLFILE,    0                },
  { "TFS", MSG_TRUNCFILE,   0                },
  { "LOK", MSG_LOCAL,       FLAG_LOCAL       },
  { "RRQ", MSG_RECEIPTREQ,  FLAG_RREQ        },
  { "CFM", MSG_CONFIRMREQ,  0                },
  { "FPU", MSG_FPU,         0                },
  { "",    MSG_INTRANSIT,   FLAG_INTRANSIT   },
  { "",    MSG_ORPHAN,      FLAG_ORPHAN      }, 
  { NULL,  0,               0                } };

void MakeFidoDate(time_t tim, char *dest)
{
   struct tm *tp;
   time_t t;
   char *monthnames[]={"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec","???"};

   t=tim;
   tp=localtime(&t);

   sprintf(dest,"%02d %s %02d  %02d:%02d:%02d",
      tp->tm_mday,
      monthnames[tp->tm_mon],
      tp->tm_year % 100,
      tp->tm_hour,
      tp->tm_min,
      tp->tm_sec);
}

void makekludge(int fh,char *pre,char *data,ulong len)
{
   char *buf = (char*)malloc(strlen(pre)+len+10);

   strcpy(buf,pre);
   if(len && data)
     strncpy(&buf[strlen(buf)],data,len+1);
    
   strcat(buf,"\x0d");
   write(fh, buf, strlen(buf));

   free(buf);
}


void extract_and_makekludge(long hKludges, int nKludge, int fh, char *prefix)
{
   char *pBuf;
   int BufLen;
   int n;
   int  nCount = _bfs_CountKludges(hKludges, nKludge);
   
   for(n=0; n < nCount; n++)
   {
			_bfs_ReadKludge(hKludges, nKludge, n, (const void **) &pBuf, (int *)&BufLen);
			makekludge(fh, prefix, pBuf, BufLen);
   }
}

int copymessage(char *name, char* from_dir, char* to_dir)
{
  char from[256], to[256];
  char buf[200];
  StoredMsg msg = {0};
 
  sprintf(from, "%s/%s", from_dir, name);
  int fh_from = open(from, O_RDONLY);
  if(fh_from < 0)
  {
    printf("Couldn't open file %s for reading\n", from);
    return -1; 
  }
  
  sprintf(to, "%s/%s", to_dir, name);

  int fh_to = open(to, O_WRONLY|O_CREAT|O_EXCL);
  if(fh_to < 0)
  {
    close(fh_from);
    printf("Couldn't open file %s for writing\n", to);
    return -1; 
  }
  
  struct Node4D n4d;
 
  _bfs_ReadAttr(from, MSG_ATTR_OADDR, buf, 200);

  if(Parse4D((unsigned char*)buf,&n4d))
  {
    msg.OrigZone=n4d.Zone;
    msg.OrigNet=n4d.Net;
    msg.OrigNode=n4d.Node;
    msg.OrigPoint=n4d.Point;
  }

  _bfs_ReadAttr(from, MSG_ATTR_DADDR, buf, 200);

  if(Parse4D((unsigned char*)buf,&n4d))
  {
    msg.DestZone=n4d.Zone;
    msg.DestNet=n4d.Net;
    msg.DestNode=n4d.Node;
    msg.DestPoint=n4d.Point;
  }

  long lo,ld;
  _bfs_ReadAttr(from, MSG_ATTR_FROM, buf, 200);
       
	ld = 36;
	lo = strlen(buf);
  _bfs_ConvertFromUTF8(MS_DOS_866_CONVERSION, buf, &lo , (char*)msg.From, &ld);
  msg.From[ld]=0;


  _bfs_ReadAttr(from, MSG_ATTR_TO, buf, 200);
 	ld = 36;
	lo = strlen(buf);
  _bfs_ConvertFromUTF8(MS_DOS_866_CONVERSION, buf, &lo , (char*)msg.To, &ld);
  msg.To[ld]=0;
            
  _bfs_ReadAttr(from, MSG_ATTR_SUBJECT, buf, 200);
 	ld = 70;
	lo = strlen(buf);
  _bfs_ConvertFromUTF8(MS_DOS_866_CONVERSION, buf, &lo , (char*)msg.Subject, &ld);
  msg.Subject[ld]=0;
  
   time_t tm;          
  _bfs_ReadAttr(from, MSG_ATTR_DATE_WRITTEN, &tm, sizeof(time_t));

  MakeFidoDate(tm, (char*)msg.DateTime);
 
  long Attribute;
  _bfs_ReadAttr(from, MSG_ATTR_ATTRIBUTES, &Attribute, sizeof(long));
  long hKludges;
  _bfs_ReadAttrKludges(from, KLUDGES_ATTR, &hKludges);

  char *pBuf;
  int BufLen;
  char flagsbuf[200];
  
  _bfs_ReadKludge(hKludges, KLUDGES_ATTR_FLAGS, 0, (const void **)&pBuf, &BufLen);
  if(pBuf)
  {
   	strcpy(flagsbuf,"\x01" "FLAGS: ");
   	strncpy(&flagsbuf[8],pBuf,BufLen+1);
  }

  for(int c=0;flagarray[c].name;c++)
    if(Attribute & flagarray[c].jamflagbit)
    {
      if(flagarray[c].fidoflagbit)
      {
        msg.Attr |= flagarray[c].fidoflagbit;
      }
      else if(flagarray[c].name[0] && strlen(flagsbuf)<90)
      {
        if(flagsbuf[0]==0)
        	strcpy(flagsbuf,"\x01" "FLAGS: ");
        else
          strcat(flagsbuf," ");

        strcat(flagsbuf,flagarray[c].name);
      }
    }

  if(flagsbuf[0])
  {
    strcat(flagsbuf,"\x0d");
    //mmAddLine(mm,flagsbuf);
    write(fh_to, flagsbuf, strlen(flagsbuf));
  }

  //oldattr = mm->Attr;
  msg.Attr = msg.Attr & (FLAG_PVT|FLAG_CRASH|FLAG_FILEATTACH|FLAG_FILEREQ|FLAG_RREQ|FLAG_IRRR|FLAG_AUDIT|FLAG_HOLD);
    
  write(fh_to, &msg, sizeof(StoredMsg));
  
  extract_and_makekludge(hKludges, KLUDGES_ATTR_MSGID, fh_to, "\x01" "MSGID: ");
  extract_and_makekludge(hKludges, KLUDGES_ATTR_REPLY, fh_to, "\x01" "REPLY: ");
  extract_and_makekludge(hKludges, KLUDGES_ATTR_FTS, fh_to, "\x01");

  struct stat st;
  fstat(fh_from, &st);
  
  void *buffer = malloc(st.st_size);
  read(fh_from, buffer, st.st_size);
  write(fh_to, buffer, st.st_size);
  free(buffer);

  extract_and_makekludge(hKludges, KLUDGES_ATTR_SEENBY, fh_to, "SEEN-BY:");
  extract_and_makekludge(hKludges, KLUDGES_ATTR_PATH, fh_to, "\x01" "PATH: ");
         
  _bfs_ReleaseKludges(hKludges);            
  
  close(fh_from);
  close(fh_to);
  printf("%s - OK\n", name);
  return 0;
}

int main(int argc, char *argv[])
{
  printf("bfs2msg converter v."VERSION"\n");
  if(argc < 3)
  {
    printf("syntax: bfs2msg <fromdir> <todir>\n");
    return -1;
  }

  if(mkdir(argv[2], 0) && errno != EEXIST)
  {
    printf("Couldn't create the directory %s\n", argv[2]);
    return -1;      
  }
 
  DIR *dp; 
  struct dirent *ep; 

  dp = opendir (argv[1]); 
  if (dp != NULL) 
  {
        
    while (ep = readdir (dp))
    {
      if(!strcmp(ep->d_name, ".") || !strcmp(ep->d_name, ".."))
        continue;
      copymessage(ep->d_name, argv[1], argv[2]);
        
    }
    (void) closedir (dp); 
  } 
  else 
    printf("Couldn't open the directory %s\n", argv[1]); 
  
  return 0;
}